import { Entity, Multi, Single, System, World, read, listenModify, ModifyEvent, write, ArrMulti } from "pi_ucore/util/ecs";
import { DataC, DataState, StateC } from "../component/data";
import { GameState, GlobalState } from "../single/global_state";
import { FrameTime, Timeout, Timer } from "pi_scene/scene/single/global";
import { MoveC } from "pi_scene/scene/component/common";
import { AnimList, AnimTime, AnimType } from "../single/anim_list";
import { Coordinate, groupCoordinates } from "../utils/base";
import { DataType, TypeC } from "../component/type";
import { BombType } from "../single/bomb_list";
import { BombProgress } from "../single/bomb_progress";
import { BombExpandTime } from "../xls/base/base.struct";

/**
 * 动画管理系统
 * 
 * 职责：
 * 1. 管理游戏内所有动画的播放时序和状态转换
 * 2. 处理炸弹爆炸的分步动画效果
 * 3. 协调动画播放与游戏状态的同步
 * 4. 管理元素移动后的回弹动画效果
 * 
 * 主要功能：
 * - 通过AnimList维护动画队列，按帧时间调度动画播放
 * - 处理合成炸弹的多阶段爆炸效果（包括火箭爆炸和区域分组爆炸）
 * - 监听游戏状态变化并触发相应动画
 * - 协调爆炸效果与游戏实体状态的更新
 */
export class AnimSys implements System {

    /**
     * 系统初始化方法（预留接口）
     * @param _world ECS世界对象
     */
    setup(_world: World) {
    }

    /**
     * 主运行逻辑 - 每帧驱动动画系统
     * @param read 包含[动画列表, 帧时间, 全局状态]的元组
     * @param write 写入器（本系统未使用）
     * 
     * 处理流程：
     * 1. 遍历当前所有动画队列
     * 2. 根据帧时间判断当前应播放的动画
     * 3. 更新游戏状态为播放动画中
     * 4. 当所有动画播放完成时切换为动画结束状态
     */
    @read(AnimList, FrameTime, GlobalState)
    run(read: [Single<AnimList>, Single<FrameTime>, Single<GlobalState>], write: any) {
        let animList = read[0].data;
        let frame = read[1].data;
        let globalState = read[2].data;
        let now = frame.curTime();
        for (let [k, arr] of animList.animList) {
            let overTime = animList.overTimeList.get(k) || 0;
            if (overTime > now) continue;

            let animData = arr.shift();
            overTime = now + animData.time

            if (arr.length === 0) {
                // 动画播完 删除
                animList.animList.delete(k);
                animList.overTimeList.delete(k);
            } else {
                animList.animList.set(k, arr);
                animList.overTimeList.set(k, overTime)
            }
            animList.overTime = Math.max(animList.overTime, overTime)
            animList.cur = animData;
            if (globalState.gameState !== GameState.PlayAnim) {
                globalState.gameState = GameState.PlayAnim;
            }
            return;
        }
        if (animList.overTime > now) return;
        if (!animList.animList.size && globalState.gameState === GameState.PlayAnim && !globalState.waitSecondPart) {
            globalState.gameState = GameState.AnimEnd;
            return;
        }
    }

    /**
     * 监听动画修改事件 - 处理爆炸动画逻辑
     * @param e 修改事件对象
     * @param read 包含[动画列表, 炸弹进度, 超时控制, 数据组件, 状态组件, 类型组件, 帧时间]的元组
     * 
     * 处理逻辑：
     * 1. 只处理爆炸类型动画（AnimType.Exploding）
     * 2. 初始化爆炸进度状态
     * 3. 处理火箭爆炸的特殊逻辑
     * 4. 分组处理爆炸区域坐标
     * 5. 设置分步爆炸定时器
     */
    @listenModify(AnimList)
    @read(AnimList, BombProgress, Timeout, [Entity, DataC], [Entity, StateC], [Entity, TypeC], FrameTime)
    listenAnim(e: ModifyEvent, read: [Single<AnimList>, Single<BombProgress>, Single<Timeout>, ArrMulti<Entity, DataC>, ArrMulti<Entity, StateC>, ArrMulti<Entity, TypeC>, Single<FrameTime>]) {
        let animList = read[0].data;
        let progress = read[1].data;
        let timeout = read[2].data;
        let dataC = read[3];
        let stateC = read[4];
        let typeC = read[5];
        let frame = read[6].data;
        let anim = animList.cur;
        console.log('播动画了', AnimType[anim.type], BombType[anim.arg], anim);
        // 只处理合成炸弹的引爆
        if (anim.type !== AnimType.Exploding || !anim.arg || anim.arg < BombType.NewCrossBomb) return;

        progress.isExploded = false; // 重置状态
        let posGroups = [], i = 0;
        let firstTime = Math.min(anim.time, AnimTime.Exploding); // 第一圈爆炸的时间
        // 超级雷的第一段爆炸没有格子
        if (!anim.posList.length) {
            progress.curEntity = anim.entityKey;
            progress.isExploded = true;
            return;
        }
        // 第一段爆炸的目标
        if (!anim.rocketEndPos && !anim.otherEndPos) {
            posGroups[0] = anim.posList;
        }
        let groupNum = 0; // 爆炸区域分组数量
        // 火箭发出的随机目标 直接一起爆炸
        if (anim.rocketEndPos?.length) {
            let rocketPosList = [];
            for (let [x, y] of anim.rocketEndPos) {
                let ind = anim.posList.findIndex(v => v[0] === x && v[1] === y)
                if (ind < 0) continue;
                rocketPosList.push(anim.posList[ind]);
            }
            groupNum++;
            this.step([rocketPosList], 0, timeout, anim.startPos[0], progress, anim.time, anim.entityKey, stateC, typeC, dataC);
        }
        // 飞行后第二段爆炸的多区域目标
        if (anim.otherEndPos?.length) {
            for (let list of anim.otherEndPos) {
                posGroups[i] = [];
                for (let [x, y] of list) {
                    let ind = anim.posList.findIndex(v => v[0] === x && v[1] === y)
                    if (ind < 0) continue;
                    posGroups[i].push(anim.posList[ind]);
                }
                i++;
            }
        }

        // 分区域 各自爆炸
        for (let posList of posGroups) {
            if (posList.length === 0) continue;
            let groups = groupCoordinates(posList, posList[0])
            groupNum++;
            console.log('坐标分组', groups, groupNum);

            // 每个区域增加一帧，避免重叠
            this.step(groups, 0, timeout, anim.startPos[0], progress, firstTime + frame.frameInterval * (groupNum - 1), anim.entityKey, stateC, typeC, dataC);
        }
        progress.groupMap.set(anim.entityKey, groupNum)
    }

    /**
     * 监听状态变化 - 处理元素移动后的回弹效果
     * @param e 修改事件对象
     * @param read 包含[全局状态, 数据组件, 状态组件]的元组
     * @param write 包含[超时控制, 移动组件]的元组
     * 
     * 触发条件：
     * 当元素状态从Moving变为Normal时，添加回弹动画
     */
    @listenModify([Entity, StateC])
    @read(GlobalState, [Entity, DataC], [Entity, StateC])
    @write(Timeout, [Entity, MoveC])
    listenState(e: ModifyEvent, read: [Single<GlobalState>, ArrMulti<Entity, DataC>, ArrMulti<Entity, StateC>],
        write: [Single<Timeout>, Multi<Entity, MoveC>]) {
        let dataC = read[1];
        let stateC = read[2];
        let timeout = write[0].data;
        let moveC = write[1];
        let state = stateC.get(e.id);
        let data = dataC.get(e.id);
        // 移动后停止 播回弹效果
        if (state.value === DataState.Normal && state.oldValue === DataState.Moving) {
            let [x, y] = data.pos;
            moveC.get(e.id).add([x, y - 0.1, 0], 80);
            // console.log('开始回弹效果', data.pos, e.id);
            timeout.value = new Timer(80, () => {
                moveC.get(e.id)?.add([x, y, 0], 48);
            })
        }
    }

    /**
     * 分步执行爆炸效果
     * @param groups 爆炸坐标分组数组
     * @param count 当前执行的分组索引
     * @param timeout 定时器控制对象
     * @param startPos 爆炸起始位置[x,y]
     * @param progress 炸弹进度对象
     * @param time 当前步骤的持续时间
     * @param entity 当前爆炸实体ID
     * @param stateC 状态组件集合
     * @param typeC 类型组件集合
     * @param dataC 数据组件集合
     * 
     * 执行逻辑：
     * 1. 按分组顺序设置定时爆炸
     * 2. 每个分组爆炸后递归调用处理下一个分组
     * 3. 当所有分组处理完成后更新爆炸状态
     */
    step(groups: Coordinate[][], count: number, timeout: Timeout, startPos: number[], progress: BombProgress, time: number, entity: number, stateC, typeC, dataC) {
        let list = groups[count] || [];
        if (count >= groups.length) {
            let groupNum = progress.groupMap.get(entity);
            groupNum--;
            if (groupNum === 0) {
                progress.groupMap.delete(entity);
                progress.curEntity = entity;
                progress.isExploded = true;
            } else {
                progress.groupMap.set(entity, groupNum);
            }
            return;
        }
        timeout.value = new Timer(time, () => {
            for (let i = 0; i < list.length; i++) {
                let [x, y, id] = list[i];
                this.bombOver(id, stateC, typeC, dataC, startPos);
            }
            this.step(groups, count + 1, timeout, startPos, progress, time ? BombExpandTime : 0, entity, stateC, typeC, dataC)
        })
    }

    /**
     * 处理单个爆炸完成逻辑
     * @param id 爆炸实体ID
     * @param stateC 状态组件集合
     * @param typeC 类型组件集合
     * @param dataC 数据组件集合
     * @param startPos 爆炸起始位置
     * 
     * 特殊处理：
     * - 奖励物品不触发爆炸状态变化
     * - 合成炸弹在非起始位置不立即改变状态
     */
    bombOver(id: number, stateC, typeC, dataC, startPos: number[]) {
        let type = typeC.get(id);
        // 奖励物品不能在此处爆炸
        if (!type || type.value === DataType.AwardData) return;
        let pos = dataC.get(id).pos
        // 引爆其他合成炸弹时，不在此时改状态
        if (type.value === DataType.SynthesisBomb && (pos[0] !== startPos[0] || pos[1] !== startPos[1])) return;
        // 修改状态不能依赖是否还有次数
        if (stateC.get(id).value !== DataState.Exploded) {
            if (type.needBombTimes > 0) type.needBombTimes--;
            stateC.get(id).value = DataState.Exploded;
        }
    }

}