import { Entity, ArrMulti, Single, System, World, read, listenModify, ModifyEvent, Multi, listenCreate } from "pi_ucore/util/ecs";
import { DataC, DataState, DirectionType, LayerType, StateC } from "../component/data";
import { GameState, GlobalState } from "../single/global_state";
import { MoveC } from "pi_scene/scene/component/common";
import { createOneData } from "../utils/init_match";
import { DataType } from "../component/type";
import { GoalC, GoalType } from "../component/goal";
import { Timeout, Timer } from "pi_scene/scene/single/global";
import { Calc } from "pi_common_tools/util/calc";
import { AnimList } from "../single/anim_list";
import { getGrid, setGrid } from "../utils/base";

/**
 * 掉落系统 - 管理游戏元素的掉落逻辑和特殊装置处理
 * 
 * 主要职责：
 * 1. 处理常规元素的下落逻辑（直线、斜向掉落）
 * 2. 管理发雷装置的特殊物品生成逻辑
 * 3. 处理元素掉落时的状态转换和碰撞检测
 * 4. 维护掉落过程中的游戏状态
 * 
 * 系统特性：
 * - 采用从左到右、从上到下的扫描顺序处理掉落
 * - 支持多种掉落方向（下、左下、右下）
 * - 实现发雷装置的物品生成和投放逻辑
 * 
 * 依赖组件：
 * - 全局状态(GlobalState)
 * - 移动组件(MoveC)
 * - 数据组件(DataC)
 * - 目标组件(GoalC)
 */
/** 正常速度移动时间配置 [直线移动时间(ms), 斜线移动时间(ms)] */
let MoveTimes = [80, 112];
/** 慢速移动时间配置 [直线移动时间(ms), 斜线移动时间(ms)] */
let SlowMoveTimes = [200, 300];
export class DropSys implements System {
    /** 标记是否需要处理掉落逻辑 */
    dirty: boolean = false;
    /** 发雷装置监听坐标映射表 [装置实体ID, 监听坐标[x,y]] */
    listenMap: Map<number, number[]> = new Map();

    /** 系统初始化方法（ECS生命周期方法） */
    setup(_world: World) {
    }

    /**
     * 主运行逻辑 - 处理每帧的掉落检测和移动控制
     * @param read 只读数据集合 [
     *   GlobalState,   // 全局游戏状态
     *   Calc,          // 计算工具
     *   StateC[],      // 实体状态组件
     *   MoveC[],       // 移动控制组件
     *   DataC[],       // 实体数据组件
     *   GoalC[],       // 目标组件
     *   Timeout,       // 定时器
     *   AnimList       // 动画列表
     * ]
     */
    @read(GlobalState, Calc, [Entity, StateC], [Entity, MoveC], [Entity, DataC], [Entity, GoalC], Timeout, AnimList)
    run(read: [Single<GlobalState>, Single<Calc>, ArrMulti<Entity, StateC>, Multi<Entity, MoveC>, ArrMulti<Entity, DataC>, ArrMulti<Entity, GoalC>, Single<Timeout>
        , Single<AnimList>], write: any) {
        let globalState = read[0].data;
        let calc = read[1].data;
        let stateC = read[2];
        let moveC = read[3];
        let dataC = read[4];
        let goalC = read[5];
        let timeout = read[6].data;
        let animList = read[7].data;
        // 动画未播完 第二段还未爆炸 不能掉落
        if (globalState.gameState === GameState.PlayAnim || globalState.waitSecondPart) return;
        if (!this.dirty) {
            if (globalState.gameState === GameState.AnimEnd) globalState.gameState = GameState.DropEnd;
            return;
        }

        let indexs = globalState.areaIndexs.get(globalState.curArea);
        let [w, h] = globalState.areaMap.get(globalState.curArea);
        this.sendBomb(write, indexs, globalState.curArea, w, h, timeout, goalC, stateC, dataC, moveC); // 发雷装置 补充物品

        let dropFlag = false;
        let list = []; // 待清除的格子
        // 从左往右 从上往下 依次检查需要移动的格子 一定要全扫描
        for (let x = 0; x < w; x++) {
            for (let y = h - 1; y >= 0; y--) {
                let id = getGrid(indexs, LayerType.Lock, x, y);
                if (id) continue;

                id = getGrid(indexs, LayerType.Data, x, y);
                if (!id) {
                    // 最顶行缺失 补物品
                    if (y === h - 1) {
                        id = this.addData(write, x, y + 1, globalState.curArea);
                        this.drop(id, x, y + 1, stateC.get(id), dataC.get(id), moveC, DirectionType.Down, indexs, MoveTimes);
                        dropFlag = true;
                        break;
                    }
                    // 可掉落物品位置
                    else if (globalState.mapCfg.dropPos.length && globalState.mapCfg.dropPos?.find(v => v[0] === x && v[1] === y)) {
                        id = this.addData(write, x, y + 1, globalState.curArea);
                        this.drop(id, x, y + 1, stateC.get(id), dataC.get(id), moveC, DirectionType.Down, indexs, MoveTimes);
                        dropFlag = true;
                        // console.log("----我要掉落了---", x, y, id)
                        break;
                    }
                    else continue;
                }
                let state = stateC.get(id);
                let move = moveC.get(id);
                // 正在移动 跳过
                if (move.isMoving() || move.arr.length > 0) {
                    dropFlag = true;
                    continue;
                }
                // 已到最后一行 切换状态
                if (y === 0) {
                    if (state.value !== DataState.Normal) state.value = DataState.Normal;
                    continue;
                }
                // 下一行被锁
                if (getGrid(indexs, LayerType.Lock, x, y - 1)) {
                    let leftFg = this.checkLeftDrop(x, y, indexs, goalC);
                    let rightFg = this.checkRightDrop(x, y, w, indexs, goalC);
                    // 不可掉落 切换状态
                    if (!leftFg && !rightFg) {
                        if (state.value !== DataState.Normal) state.value = DataState.Normal;
                        continue;
                    }
                    let direction = leftFg ? DirectionType.Left : DirectionType.Right
                    list.push([x, y]);
                    this.drop(id, x, y, state, dataC.get(id), moveC, direction, indexs, MoveTimes);
                    dropFlag = true;
                    break;
                }

                // 下一行为空 或 下一行正在移动
                let downId = getGrid(indexs, LayerType.Data, x, y - 1);
                let downFg = !downId || (moveC.get(downId).isMoving() && dataC.get(downId).pos[1] !== y - 1);
                // 下一行是否停止移动
                let downStopFg = downId && stateC.get(downId).value === DataState.Normal;
                let leftFg = downStopFg && this.checkLeftDrop(x, y, indexs, goalC);
                let rightFg = downStopFg && this.checkRightDrop(x, y, w, indexs, goalC);

                if (!downFg && !leftFg && !rightFg) {
                    // 下一行已停止移动 先处理下一行
                    if (!moveC.get(downId).isMoving() && stateC.get(downId).value === DataState.Moving) continue;
                    // 不再需要移动 状态改为普通
                    if (state.value === DataState.Moving) state.value = DataState.Normal;
                    continue;
                }
                let direction = downFg ? DirectionType.Down : (leftFg ? DirectionType.Left : DirectionType.Right)
                list.push([x, y]);
                this.drop(id, x, y, state, dataC.get(id), moveC, direction, indexs, MoveTimes);
                dropFlag = true;
                break;
            }
        }
        for (let [x, y] of list) {
            setGrid(indexs, LayerType.Data, x, y, 0)
        }
        if (!dropFlag) {
            this.dirty = false;
            globalState.gameState = GameState.DropEnd;
        }
        else if (globalState.gameState !== GameState.Droping) {
            globalState.gameState = GameState.Droping;
        }
    }

    /**
     * 监听实体状态变化 - 处理掉落触发和发雷装置计数
     * @param e 修改事件对象
     * @param read 只读数据集合 [
     *   GlobalState,   // 全局游戏状态
     *   DataC[],       // 实体数据组件
     *   StateC[],      // 实体状态组件
     *   GoalC[]        // 目标组件
     * ]
     */
    @listenModify([Entity, StateC])
    @read(GlobalState, [Entity, DataC], [Entity, StateC], [Entity, GoalC])
    listenSpeed(e: ModifyEvent, read: [Single<GlobalState>, ArrMulti<Entity, DataC>, ArrMulti<Entity, StateC>, Multi<Entity, GoalC>], write: any) {
        let globalState = read[0].data;
        let dataC = read[1];
        let stateC = read[2];
        let goalC = read[3];
        let state = stateC.get(e.id);
        if (state.value === DataState.Exploded || state.value === DataState.Normal) { // 已爆炸
            this.dirty = true;
        }
        if (state.value === DataState.Moving) { // 移动
            let data = dataC.get(e.id);
            let [x, y] = data.pos;
            for (let [id, [x1, y1]] of this.listenMap) {
                if (x === x1 && y === y1) {
                    goalC.get(id).dataCount++;
                    console.log('发雷装置', e.id, goalC.get(id).dataCount, [x1, y1]);
                }
            }
        }
    }

    /**
     * 监听目标组件创建 - 初始化发雷装置监听位置
     * @param e 创建事件对象
     * @param read 只读数据集合 [
     *   GoalC[],  // 目标组件
     *   DataC[]   // 数据组件
     * ]
     */
    @listenCreate([Entity, GoalC])
    @read([Entity, GoalC], [Entity, DataC])
    listenCreateType(e: ModifyEvent, read: [Multi<Entity, GoalC>, ArrMulti<Entity, DataC>,], write: any) {
        let goalC = read[0];
        let dataC = read[1];
        let goal = goalC.get(e.id);
        let data = dataC.get(e.id);
        if (goal.type === GoalType.SendBomb) {
            let direction = goal.condition.direction;
            let [x, y] = data.pos;
            let x1 = x, y1 = y;
            if (direction === DirectionType.Up) y1 = y + 1;
            if (direction === DirectionType.Down) y1 = y - 1;
            if (direction === DirectionType.Left) x1 = x - 1;
            if (direction === DirectionType.Right) x1 = x + 1;
            this.listenMap.set(e.id, [x1, y1]);
        }
    }


    /**
     * 执行元素掉落操作
     * @param id 实体ID
     * @param x 当前X坐标
     * @param y 当前Y坐标
     * @param state 状态组件
     * @param data 数据组件
     * @param moveC 移动控制组件
     * @param direction 掉落方向 
     * @param indexs 地图索引数据
     */
    drop(
        id: number,
        x: number,
        y: number,
        state: StateC,
        data: DataC,
        moveC: Multi<Entity, MoveC>,
        direction: DirectionType,
        indexs: number[][][],
        moveTimes: number[]
    ) {
        if (direction === DirectionType.Down) { // 下
            moveC.get(id).add([x, y - 1, 0], moveTimes[0])
            setGrid(indexs, LayerType.Data, x, y - 1, id)
            data.pos = [x, y - 1];
        }
        else if (direction === DirectionType.SlowDown) { // 缓慢向下
            moveC.get(id).add([x, y - 0.01, 0], 800)
            moveC.get(id).add([x, y - 1, 0], moveTimes[0])
            setGrid(indexs, LayerType.Data, x, y - 1, id)
            data.pos = [x, y - 1];
        }
        else if (direction === DirectionType.Left) { // 左下
            moveC.get(id).add([x - 1, y - 1, 0], moveTimes[1])
            setGrid(indexs, LayerType.Data, x - 1, y - 1, id)
            data.pos = [x - 1, y - 1];
        }
        else if (direction === DirectionType.Right) { // 右下
            moveC.get(id).add([x + 1, y - 1, 0], moveTimes[1])
            setGrid(indexs, LayerType.Data, x + 1, y - 1, id)
            data.pos = [x + 1, y - 1];
        }
        else if (direction === DirectionType.Other) { // 其他
            moveC.get(id).add([x, y, 0], moveTimes[0])
            setGrid(indexs, LayerType.Data, x, y, id)
            data.pos = [x, y];
        }
        // 记录移动中状态
        state.value = DataState.Moving;
        // console.log('开始掉落', [x, y], id, DirectionType[direction], JSON.stringify(indexs[LayerType.Data]));
    }

    /**
     * 创建新物品并添加到指定位置
     * @param world ECS世界对象
     * @param x X坐标
     * @param y Y坐标
     * @param area 区域ID
     * @returns 新创建的实体ID
     */
    addData(world: World, x: number, y: number, area: number) {
        return createOneData(world, [x, y], area, LayerType.Data);
    }

    /**
     * 检查是否可以进行左下方掉落
     * @param x 当前X坐标
     * @param y 当前Y坐标
     * @param indexs 地图索引数据
     * @param goalC 目标组件
     * @returns 是否可以左下方掉落
     */
    checkLeftDrop(x, y, indexs, goalC) {
        let flag = x > 0;
        if (!flag) return flag;
        let leftId = !!getGrid(indexs, LayerType.Lock, x - 1, y);
        // 左侧被锁
        flag = leftId;
        // 自身和左侧上方同时被锁
        if (!flag) flag = this.checkUpLock(x, y + 1, indexs, goalC, true) && this.checkUpLock(x - 1, y, indexs, goalC)

        if (!flag) return flag;
        // 左下为空
        flag = !getGrid(indexs, LayerType.Lock, x - 1, y - 1) && !getGrid(indexs, LayerType.Data, x - 1, y - 1);

        return flag;
    }

    /**
     * 检查是否可以进行右下方掉落
     * @param x 当前X坐标
     * @param y 当前Y坐标
     * @param w 地图宽度
     * @param indexs 地图索引数据
     * @param goalC 目标组件
     * @returns 是否可以右下方掉落
     */
    checkRightDrop(x, y, w, indexs, goalC) {
        let flag = x + 1 < w;
        if (!flag) return flag;
        let rightId = !!getGrid(indexs, LayerType.Lock, x + 1, y);
        // 右侧被锁
        flag = rightId;
        // 自身和右侧上方同时被锁
        if (!flag) flag = this.checkUpLock(x, y + 1, indexs, goalC, true) && this.checkUpLock(x + 1, y, indexs, goalC)

        if (!flag) return flag;
        // 右下为空
        flag = !getGrid(indexs, LayerType.Lock, x + 1, y - 1) && !getGrid(indexs, LayerType.Data, x + 1, y - 1);

        return flag;
    }

    /**
     * 检查指定坐标上方是否存在锁定状态
     * @param x 检查的X坐标
     * @param y 起始Y坐标
     * @param indexs 地图索引数据
     * @param goalC 目标组件
     * @param isSelf 是否检查自身上方
     * @returns 上方是否存在锁定
     */
    checkUpLock(x, y, indexs, goalC, isSelf = false) {
        let arr = indexs[LayerType.Lock][x];
        for (let i = y; i < arr.length; i++) {
            // 自身上方有锁就算
            if (isSelf && arr[i]) return true;
            // 非自身上方 发雷装置不算锁
            if (arr[i]) return true;
            if (getGrid(indexs, LayerType.Data, x, i)) return false;
        }
        return false;
    }

    /**
     * 处理发雷装置的物品生成逻辑
     * @param world ECS世界对象
     * @param indexs 地图索引数据
     * @param area 区域ID
     * @param w 地图宽度
     * @param h 地图高度
     * @param timeout 定时器
     * @param goalC 目标组件
     * @param stateC 状态组件
     * @param dataC 数据组件
     * @param moveC 移动控制组件
     */
    sendBomb(world, indexs: number[][][], area: number, w: number, h: number, timeout: Timeout, goalC, stateC, dataC, moveC) {
        // 从左往右 从上往下
        for (let x = 0; x < w; x++) {
            for (let y = 0; y < h; y++) {
                let id = getGrid(indexs, LayerType.Lock, x, y);
                if (!id) continue;
                let goal = goalC.get(id);
                if (goal?.type !== GoalType.SendBomb) continue;

                let list = goal.condition.list;
                let [x1, y1] = this.listenMap.get(id);
                id = getGrid(indexs, LayerType.Data, x1, y1);
                // 可以弹出物品
                if (!id && goal.dataCount === goal.goalCount) {
                    let val = list.shift();
                    if (val) {
                        if (val > DataType.TimeAwardData) {
                            id = createOneData(world, [x, y], area, LayerType.Data, [1, 1], 1, DataType.TimeAwardData, val);
                        } else if (val > DataType.SynthesisBomb) {
                            id = createOneData(world, [x, y], area, LayerType.Data, [1, 1], 1, DataType.SynthesisBomb, val);
                        } else {
                            id = createOneData(world, [x, y], area, LayerType.Data, [1, 1], 1, val);
                        }
                        console.log('发雷装置弹出物品', id, val, list);
                        this.drop(id, x1, y1, stateC, dataC.get(id), moveC, DirectionType.Other, indexs, SlowMoveTimes)
                        goal.dataCount = 0;
                    }
                }
            }
        }
    }
}
