import { Entity, Multi, Single, System, World, read, listenModify, ModifyEvent, ArrMulti } from "pi_ucore/util/ecs";
import { DataC, DataState, LayerType, SizeC, StateC } from "../component/data";
import { GlobalState } from "../single/global_state";
import { createOneData, destroyOneData } from "../utils/init_match";
import { BombList, BombType } from "../single/bomb_list";
import { DataType, NormalLockType, TypeC } from "../component/type";
import { GoalC, GoalType } from "../component/goal";
import { AnimList, AnimTime, AnimType } from "../single/anim_list";
import { Collect } from "../single/collect";
import { Timeout, Timer } from "pi_scene/scene/single/global";
import { AroundPosFour, Coordinate, getGrid, groupAdjacentCoordinates, isAllHaystack, isHaveWater, setGrid } from "../utils/base";
import { BombProgress } from "../single/bomb_progress";

/**
 * 爆炸系统 - 处理爆炸效果及连锁反应
 * 
 * 主要职责：
 * 1. 处理爆炸后的格子状态更新
 * 2. 管理炸弹连锁反应（合成炸弹、超级炸弹）
 * 3. 处理草垛层数逻辑和隐藏物品显示
 * 4. 水域扩展逻辑
 * 5. 收集品统计和动画触发
 * 
 * 系统依赖：
 * - 需要与BombProgress协同工作确保爆炸顺序
 * - 依赖AnimList处理爆炸相关动画
 */
export class ExplodeSys implements System {
    /**
     * 待处理的爆炸实体数组
     * [实体ID列表] 记录需要处理爆炸效果的实体
     */
    dirtyList: number[] = [];

    // 定义初始化方法（可选）
    setup(_world: World) {
    }

    @read(GlobalState, [Entity, DataC], [Entity, StateC], [Entity, TypeC], [Entity, SizeC], AnimList, Collect, [Entity, GoalC], BombList, BombProgress)
    /**
     * 主运行逻辑
     * @param read 只读组件集合 [
     *   GlobalState (全局状态),
     *   DataC (数据组件),
     *   StateC (状态组件),
     *   TypeC (类型组件),
     *   SizeC (尺寸组件),
     *   AnimList (动画列表),
     *   Collect (收集统计),
     *   GoalC (目标组件),
     *   BombList (炸弹列表),
     *   BombProgress (爆炸进度)
     * ]
     * @param write 可写世界实例，用于创建/销毁实体
     */
    run(
        read: [
            Single<GlobalState>,
            ArrMulti<Entity, DataC>,
            ArrMulti<Entity, StateC>,
            ArrMulti<Entity, TypeC>,
            Multi<Entity, SizeC>,
            Single<AnimList>,
            Single<Collect>,
            Multi<Entity, GoalC>,
            Single<BombList>,
            Single<BombProgress>
        ],
        write: World
    ) {
        // 空队列快速返回
        if (this.dirtyList.length === 0) return;

        let globalState = read[0].data;
        let dataC = read[1];
        let stateC = read[2];
        let typeC = read[3];
        let sizeC = read[4];
        let animList = read[5].data;
        let collect = read[6].data;
        let goalC = read[7];
        let bombList = read[8].data;
        let progress = read[9].data;
        // 爆炸全部结束再处理销毁格子
        if (!progress.isExploded) return;

        console.log('ExplodeSys dirtyList = ', this.dirtyList, globalState.cannotBombList)
        let indexs = globalState.areaIndexs.get(globalState.curArea);
        let [w, h] = globalState.areaMap.get(globalState.curArea);

        let posList: Coordinate[] = [];
        // 不能炸的物品 但可以扩展水域
        for (let id of globalState.cannotBombList) {
            let [x, y] = dataC.get(id).pos;
            if (typeC.get(id).isEmpty() || posList.find((v) => v[0] === x && v[1] === y)) continue;
            posList.push([x, y, id]);
        }
        globalState.cannotBombList = [];

        for (let id of this.dirtyList) {
            let data = dataC.get(id);
            let size = sizeC.get(id);
            let type = typeC.get(id);
            let state = stateC.get(id);
            let newId = 0;
            let canRemove = type.needBombTimes === 0; // 是否可以移除
            if (type.hideType && canRemove) {
                let val = type.hideType === BombType.SplitBomb ? DataType.SynthesisBomb : DataType.AwardData;
                newId = createOneData(write, data.pos, globalState.curArea, LayerType.Data, [1, 1], 0, val, type.hideType, type.subTypeArg);
                console.log('替换为奖励物品了', newId)
            }
            // 引爆最后一个合成雷 创建出超级雷
            if (type.value === DataType.SynthesisBomb && bombList.waitBombs.length === 0 && bombList.superBombType) {
                newId = createOneData(write, data.pos, globalState.curArea, LayerType.Data, [1, 1], 1, DataType.SynthesisBomb, bombList.superBombType);
                bombList.superBombType = null;
                console.log('替换为三级炸弹', newId, BombType[bombList.superBombType]);
            }
            let [x, y] = data.pos;
            // 不能在此处销毁草垛
            if (canRemove && data.layer !== LayerType.Lawn) destroyOneData(write, id);

            // 占地范围 移除格子
            for (let i = 0; i < size.width; i++) {
                for (let j = 0; j < size.height; j++) {
                    let x1 = x + i, y1 = y + j
                    // 不能在此处改变草垛的id
                    if (getGrid(indexs, data.layer, x1, y1) === id && canRemove && data.layer !== LayerType.Lawn) setGrid(indexs, data.layer, x1, y1, newId);
                    // 被炸弹炸到的格子才算 周围影响的不算
                    if (state.oldValue === DataState.WaitExplode) posList.push([x1, y1, id]);
                }
            }
            // 还需要被炸才能移除 修改状态
            if (!canRemove) {
                state.value = DataState.Normal;
            }
            // 处理草垛
            let synthesisBomb = false;
            if (bombList.waitSynthesisPos) {
                const index = bombList.waitSynthesisPos.findIndex(v => v[0] === x && v[1] === y);
                if (index >= 0) {
                    synthesisBomb = true;
                    bombList.waitSynthesisPos.splice(index, 1)
                }
            }
            this.haystack(write, indexs, collect, globalState, [x, y], goalC, typeC, canRemove && !type.hideType && !synthesisBomb, data.layer === LayerType.Lawn);
        }
        // 扩展水域
        this.expandWater(indexs, posList, animList, collect, w, h, typeC, stateC);

        this.dirtyList = [];
        console.log('ExplodeSys lawn indexs = ', JSON.stringify(indexs[LayerType.Lawn]));
        console.log('ExplodeSys data indexs = ', JSON.stringify(indexs[LayerType.Data]));
        console.log('ExplodeSys lock indexs = ', JSON.stringify(indexs[LayerType.Lock]));
    }

    /**
     * 监听状态变化事件
     * @param e 修改事件
     * @param read 只读组件集合 [
     *   GlobalState,
     *   DataC组件,
     *   StateC组件,
     *   TypeC组件,
     *   GoalC组件,
     *   SizeC组件,
     *   Timeout单例,
     *   AnimList单例
     * ]
     * @param write 可写世界实例
     * 
     * 主要处理：
     * 1. 美人鱼锁解锁动画
     * 2. 珍珠收集统计
     * 3. 卷帘门收起逻辑
     */
    @listenModify([Entity, StateC])
    @read(GlobalState, [Entity, DataC], [Entity, StateC], [Entity, TypeC], [Entity, GoalC], [Entity, SizeC], Timeout, AnimList)
    listenState(
        e: ModifyEvent,
        read: [
            Single<GlobalState>,
            ArrMulti<Entity, DataC>,
            ArrMulti<Entity, StateC>,
            ArrMulti<Entity, TypeC>,
            Multi<Entity, GoalC>,
            Multi<Entity, SizeC>,
            Single<Timeout>,
            Single<AnimList>
        ],
        write: World
    ) {
        let globalState = read[0].data;
        let dataC = read[1];
        let stateC = read[2];
        let typeC = read[3];
        let goalC = read[4];
        let sizeC = read[5];
        let timeout = read[6].data;
        let animList = read[7].data;
        let state = stateC.get(e.id);
        if (e.id === 1) console.log("珍珠盘更新", state.value);
        if (state.value !== DataState.Exploded) return;
        let data = dataC.get(e.id);
        let type = typeC.get(e.id);
        let goal = goalC.get(e.id);
        let size = sizeC.get(e.id);
        let indexs = globalState.areaIndexs.get(globalState.curArea);
        let [w, h] = globalState.areaMap.get(globalState.curArea);

        // 处理产出型目标（珍珠/美人鱼锁）
        if ((state.oldValue === DataState.WaitExplode /* 等待爆炸状态 */
            || state.oldValue === DataState.Normal /* 普通状态 */)
            && goal?.type === GoalType.OutPut /* 产出型目标 */) {
            if (goal.dataType === DataType.MermaidLock) {
                if (goal.dataCount < goal.goalCount && goal.dataCount + 1 >= goal.goalCount) {
                    animList.addAnimData(e.id, {
                        type: AnimType.UnlockMermaid,
                        time: AnimTime.UnlockMermaid,
                        posList: [data.pos]
                    });
                    state.value = DataState.WaitExplode;
                    timeout.value = new Timer(AnimTime.UnlockMermaid, (id) => {
                        goal.dataCount++;
                        state.value = DataState.Exploded;
                        console.warn('美人鱼收集', id, data.pos);
                    }, e.id)
                } else {
                    goal.dataCount++;
                }
            } else {
                if (goal.dataType === DataType.Pearl) {
                    console.log("=========收集珍珠===========自己", type.needBombTimes)
                }
                goal.dataCount++;
                console.log(e.id, '炸自己', goal.dataCount);
            }
        }
        // 卷起一格 往下卷
        if (goal?.type === GoalType.RollUp) {
            let time = type.needBombTimes ? AnimTime.BlindsCurlUp : 0;
            // 最后一格不需要动画
            if (time) {
                animList.addAnimData(e.id, {
                    type: AnimType.BlindsCurlUp,
                    time: AnimTime.BlindsCurlUp,
                    posList: [data.pos],
                    entityKey: e.id
                });
            }
            timeout.value = new Timer(time, (id) => {
                if (size.height) {
                    size.height--;
                    let [x, y] = data.pos;
                    setGrid(indexs, LayerType.Lock, x, y + size.height, 0);
                    console.log('卷起一格了', id, [x, y], size.height);
                }
            }, e.id)
        }
        if (!this.dirtyList.includes(e.id)) this.dirtyList.push(e.id);
    }

    /**
     * 处理草垛逻辑
     * @param world 游戏世界实例
     * @param indexs 三维索引数组 [层类型][x][y]
     * @param collect 收集统计单例
     * @param globalState 全局状态
     * @param pos 当前处理位置 [x, y]
     * @param goalC 目标组件集合
     * @param typeC 类型组件集合
     * @param exploded 是否完全爆炸
     * @param self 是否是自身爆炸
     */
    haystack(
        world: World,
        indexs: number[][][],
        collect: Collect,
        globalState: GlobalState,
        [x, y]: number[],
        goalC: Multi<Entity, GoalC>,
        typeC: ArrMulti<Entity, TypeC>,
        exploded: boolean,
        self: boolean
    ) {
        let lawnId = getGrid(indexs, LayerType.Lawn, x, y);
        if (lawnId <= 0) return;

        let lawnGoal = goalC.get(lawnId);
        let lawnType = typeC.get(lawnId);
        // 炸自身但没炸完不显示隐藏物品
        if (self && !exploded) return;

        // 其上的物品消失且物品层没有物品 则草垛可以减一层
        if (exploded && !getGrid(indexs, LayerType.Data, x, y) && !self) {
            this.delTimes(lawnType, collect);
        }
        // 其上的物品如果不能消失 则草垛也不能消失 但如果有多层可以减少一层
        else if (lawnType.needBombTimes > 1) {
            this.delTimes(lawnType, collect);
        }
        this.showData(world, indexs, lawnType, lawnGoal, globalState, [x, y], lawnId)
    }

    // 减少草垛层数
    delTimes(lawnType: TypeC, collect: Collect) {
        if (!lawnType.needBombTimes) return;
        lawnType.needBombTimes--;
        collect.addCount(DataType.Haystack, 1)
        console.log('草垛层数减一', lawnType.needBombTimes)
    }

    // 显示隐藏物品
    showData(world: World, indexs: number[][][], lawnType: TypeC, lawnGoal: GoalC, globalState: GlobalState, [x, y]: number[], lawnId: number) {
        if (lawnType.needBombTimes > 0) return;

        let val = lawnGoal.condition.list.shift();
        console.log('草垛下一个物品', val, lawnGoal.condition.list)
        if (val === DataType.Haystack) {
            lawnType.needBombTimes = lawnGoal.condition.count;
            console.log('显示出新的草垛了', [x, y])
        } else if (val) {
            let newId: number;
            if (val > DataType.TimeAwardData) {
                newId = createOneData(world, [x, y], globalState.curArea, LayerType.Data, [1, 1], 1, DataType.TimeAwardData, val);
            } else {
                newId = createOneData(world, [x, y], globalState.curArea, LayerType.Data, [1, 1], 1, val);
            }

            setGrid(indexs, LayerType.Data, x, y, newId);
            console.log('显示出草垛下的隐藏物品了', newId)
            // 再弹一个草垛出来
            this.showData(world, indexs, lawnType, lawnGoal, globalState, [x, y], lawnId);
        } else {
            // 删除草垛
            destroyOneData(world, lawnId)
            setGrid(indexs, LayerType.Lawn, x, y, 0);
            console.log('销毁草垛了', [x, y])
        }
    }

    /**
     * 扩展水域逻辑
     * @param indexs 三维索引数组 [层类型][x][y]
     * @param posList 需要检测的坐标列表（包含实体ID）
     * @param animList 动画列表单例
     * @param collect 收集统计单例
     * @param w 当前区域宽度
     * @param h 当前区域高度
     * @param typeC 类型组件集合
     * @param stateC 状态组件集合
     * 
     * 实现步骤：
     * 1. 对爆炸点进行相邻分组（四方向连接）
     * 2. 组内只要有一个坐标邻接水域即可继续处理
     * 3. 每个坐标单独检测阻挡条件（四方向是否全被阻挡）
     * 4. 未被阻挡且邻接水域的坐标标记为水域
     * 5. 触发铺草动画和统计
     */
    expandWater(
        indexs: number[][][],
        posList: Coordinate[],
        animList: AnimList,
        collect: Collect,
        w: number,
        h: number,
        typeC: ArrMulti<Entity, TypeC>,
        stateC: ArrMulti<Entity, StateC>
    ) {
        let lawnCount = 0, lawnPosList = [];
        // 将坐标按是否相邻分堆
        let groups = groupAdjacentCoordinates(posList);

        // 处理每个爆炸点组
        groups.forEach(group => {
            // 找到组内所有与水域相邻的坐标
            const waterAdjacentPositions = group.filter(([x, y]) =>
                isHaveWater(indexs, [x, y], w, h));

            // 如果组内有与水域相邻的坐标，则进行扩展
            if (waterAdjacentPositions.length > 0) {
                // 使用广度优先搜索（BFS）从水域相邻的坐标开始扩展
                const queue: Coordinate[] = [...waterAdjacentPositions];
                const visited = new Set<string>();

                while (queue.length > 0) {
                    const [x, y, id] = queue.shift()!;
                    const key = `${x},${y}`;

                    if (visited.has(key)) continue;
                    visited.add(key);

                    // 检查当前坐标是否在 posList 中
                    if (!posList.some(([px, py]) => px === x && py === y)) continue;

                    // 检查当前坐标是否可以被扩展
                    if (getGrid(indexs, LayerType.Lawn, x, y) === 0 && // 当前位置未被占领
                        !isAllHaystack(indexs, [x, y], w, h) && // 未被完全阻挡
                        x >= 0 && x < w && y >= 0 && y < h) { // 在有效范围内

                        // 执行水域扩展
                        setGrid(indexs, LayerType.Lawn, x, y, -1);
                        lawnPosList.push([x, y]);
                        lawnCount++;

                        // 特殊处理：解除冰块鱼的冻结状态
                        const entityType = typeC.get(id)?.value;
                        if (entityType === DataType.Donut) {
                            stateC.get(id).value = DataState.Normal;
                        }

                        // 将四方向相邻的坐标加入队列（仅限 posList 中的坐标）
                        for (let [i, j] of AroundPosFour) {
                            const x1 = x + i, y1 = y + j;
                            if (x1 >= 0 && x1 < w && y1 >= 0 && y1 < h) {
                                // 检查相邻坐标是否在 posList 中
                                let ind = posList.findIndex(([px, py]) => px === x1 && py === y1)
                                if (ind > -1) {
                                    queue.push(posList[ind]);
                                }
                            }
                        }
                    }
                }
            }
        });

        console.log('扩展水域', lawnCount);
        if (lawnCount) {
            animList.addAnimData(0, {
                type: AnimType.LayLawn,
                time: AnimTime.LayLawn,
                posList: lawnPosList
            });
            collect.addCount(DataType.Lawn, lawnCount);
        }
    }

}
