import { Entity, ArrMulti, Single, System, World, read, listenModify, ModifyEvent, Multi, listenCreate } from "pi_ucore/util/ecs";
import { DataC, DataState, DirectionType, LayerType, StateC } from "../component/data";
import { GameState, GlobalState } from "../single/global_state";
import { MoveC } from "pi_scene/scene/component/common";
import { createOneData } from "../utils/init_match";
import { DataType } from "../component/type";
import { GoalC, GoalType } from "../component/goal";
import { Timeout, Timer } from "pi_scene/scene/single/global";
import { Calc } from "pi_common_tools/util/calc";
import { AnimList } from "../single/anim_list";
import { BombList, BombType } from "../single/bomb_list";
import { TimeStep } from "../single/time_step";

/**
 * 定时步骤系统
 */
export class TimeStepSys implements System {
    /** 标记是否需要处理掉落逻辑 */
    dirty: boolean = false;

    /** 系统初始化方法（ECS生命周期方法） */
    setup(_world: World) {
    }

    /**
     * 主运行逻辑 - 处理每帧的掉落检测和移动控制
     * @param read 只读数据集合 [
     *   GlobalState,   // 全局游戏状态
     *   Calc,          // 计算工具
     *   StateC[],      // 实体状态组件
     *   MoveC[],       // 移动控制组件
     *   DataC[],       // 实体数据组件
     *   GoalC[],       // 目标组件
     *   Timeout,       // 定时器
     *   AnimList       // 动画列表
     * ]
     */
    @read(GlobalState, Calc, [Entity, StateC], [Entity, MoveC], [Entity, DataC], [Entity, GoalC], Timeout, AnimList)
    run(read: [Single<GlobalState>, Single<Calc>, ArrMulti<Entity, StateC>, Multi<Entity, MoveC>, ArrMulti<Entity, DataC>, ArrMulti<Entity, GoalC>, Single<Timeout>
        , Single<AnimList>], write: any) {
    }

    /**
     * 监听全局状态变化事件（主逻辑入口）
     * @param e - 修改事件对象
     * @param read - 读取的组件数据集合，包含：
     *   [0] GlobalState 全局状态
     *   [1] Calc 计算工具
     *   [2] BombList 炸弹列表
     *   [3] GoalC 目标组件
     *   [4] TypeC 类型组件
     *   [5] DataC 数据组件
     *   [6] Timeout 超时控制
     *   [7] AnimList 动画列表
     */
    @listenModify(GlobalState)
    @read(GlobalState, BombList, TimeStep)
    listenGlobalState(e: ModifyEvent, read: [Single<GlobalState>, Single<BombList>, Single<TimeStep>],
        write: any) {
        let globalState: GlobalState = read[0].data;
        let bombList = read[1].data;
        let timeStep: TimeStep = read[2].data;

        // 在掉落结束状态且没有等待中的炸弹时处理扩展逻辑
        if (globalState.gameState === GameState.DropEnd && !bombList.waitBombs.length) {
            timeStep.step += 1;
        }
    }


    // 检查定时爆炸节点
    private checkAutoBoom() {
        
    }
}
